@@ -10,7 +10,7 @@ use strict;
 use warnings;
 our ( %Config, $VERSION );
 
-$VERSION = "5.028001";
+$VERSION = "5.030000";
 
 # Skip @Config::EXPORT because it only contains %Config, which we special
 # case below as it's not a function. @Config::EXPORT won't change in the
@@ -56,11 +56,12 @@ sub import {
     return;
 }
 
-die "$0: Perl lib version (5.28.1) doesn't match executable '$^X' version ($])"
+die "$0: Perl lib version (5.30.0) doesn't match executable '$^X' version ($])"
     unless $^V;
 
-$^V eq 5.28.1
-    or die sprintf "%s: Perl lib version (5.28.1) doesn't match executable '$^X' version (%vd)", $0, $^V;
+$^V eq 5.30.0
+    or die sprintf "%s: Perl lib version (5.30.0) doesn't match executable '$^X' version (%vd)", $0, $^V;
+
 
 sub FETCH {
     my($self, $key) = @_;
@@ -83,7 +84,7 @@ sub AUTOLOAD {
 
 # tie returns the object, so the value returned to require will be true.
 tie %Config, 'Config', {
-    archlibexp => '/usr/lib/powerpc64le-linux-gnu/perl/5.28',
+    archlibexp => '/usr/lib/powerpc64le-linux-gnu/perl/5.30',
     archname => 'powerpc64le-linux-gnu-thread-multi',
     cc => 'powerpc64le-linux-gnu-gcc',
     d_readlink => 'define',
@@ -92,19 +93,19 @@ tie %Config, 'Config', {
     dlsrc => 'dl_dlopen.xs',
     dont_use_nlink => undef,
     exe_ext => '',
-    inc_version_list => '5.28.0 5.28.0/powerpc64le-linux-gnu-thread-multi',
+    inc_version_list => '',
     intsize => '4',
     ldlibpthname => 'LD_LIBRARY_PATH',
-    libpth => '/usr/local/lib /usr/lib/gcc/powerpc64le-linux-gnu/9/include-fixed /usr/include/powerpc64le-linux-gnu /usr/lib /lib/powerpc64le-linux-gnu /lib/../lib /usr/lib/powerpc64le-linux-gnu /usr/lib/../lib /lib',
+    libpth => '/usr/local/lib /usr/include/powerpc64le-linux-gnu /usr/lib /lib/powerpc64le-linux-gnu /lib/../lib /usr/lib/powerpc64le-linux-gnu /usr/lib/../lib /lib',
     osname => 'linux',
-    osvers => '4.9.0',
+    osvers => '4.19.0',
     path_sep => ':',
-    privlibexp => '/usr/share/perl/5.28',
+    privlibexp => '/usr/share/perl/5.30',
     scriptdir => '/usr/bin',
-    sitearchexp => '/usr/local/lib/powerpc64le-linux-gnu/perl/5.28.1',
-    sitelibexp => '/usr/local/share/perl/5.28.1',
+    sitearchexp => '/usr/local/lib/powerpc64le-linux-gnu/perl/5.30.0',
+    sitelibexp => '/usr/local/share/perl/5.30.0',
     so => 'so',
     useithreads => 'define',
     usevendorprefix => 'define',
-    version => '5.28.1',
+    version => '5.30.0',
 };
